/************************************************************************/
/*                                                                      */
/*   Program Name :  k5tom5                                             */
/*                                                                      */
/*     K5tom5 is a K5 to Mark5 file format converter                    */
/*                                                                      */
/*     Enviroment variables                                             */
/*           M5DIR      ---   default directory for Mark5 data          */
/*           M5VEX      ---   default directory for Mark5 vex file      */
/*                                                                      */
/*            Ver 1.21  2005-09-17                                      */
/*                                                                      */
/*      Copyright (c) 2004-2005 T.Kondo/NICT All Right Reserved         */
/*                                                                      */
/************************************************************************/
/*
     2004-11-18  -- Ver 1.00
     2004-12-28  -- Ver 1.10
     2005-01-13  -- Ver 1.11
     2005-06-17  -- Ver 1.20  update due to library restructuring
     2005-09-17  -- Ver 1.21  

2005-09-17
     use 'LL' like 0xFFFFFFFFFFFFFFFFLL to express int64_t constant 
     to avoid warning message appearing in the latest version of gcc.

   how to run
          k5tom5 jjj|yyyymmdd|yyyyddd 
                        k5name1 [k5name2 k5name3 k5name4 [m5name]] [options]
          
            or 
          k5tom5 -i -info_file|make -v vex_file [options]
                         (for information file creation only!!)

    
    where    
             jjj|yyyymmdd|yyyyddd  -- 
                     jjj      : lower 3digits of MJD 
                       or
                     yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month
                       or
                     yyyyddd  : 4 digit year, 3 digit day of year

                        of observation date
             
             k5name1 -- K5 data file name containing Ch01-04 or Ch01
             k5name2 -- K5 data file name containing Ch05-08 or Ch02
             k5name3 -- K5 data file name containing Ch09-12 or Ch03
             k5name4 -- K5 data file name containing Ch13-16 or Ch04

             m5name  -- Mark5 file name to be created

             options (any order)
               -i info_file  -- infomation file name that contains
                                  track vs channel table
                                  bit_position vs channel table
                                  group# vs channel table
                                  fanout# vs track
                                  (default "k5tom5info.txt")
                                  "-i -info_file"  means info_file named info_file creation
                                  "-i make" means default info_file creation
               -v vex_file   -- VEX file name to be read when "-i make" is set
               -sid stat_id  -- station ID when "-i make" is set
               -scan scan#   -- scan# for MODE get when "-i make" is set
                                 (default : 1)\n");
               -o m5name     -- Mark5 file name to be created
                                 (default : see below)
               -d m5dir      -- directory for Mark5 file out
                                 (default  : mark5 directory)
               -s strat_sec  -- offset time from data head (sec)
                                 (default :  0 )
               -p period     -- data period to convert (sec)
                                 (default : all)
               -mark4        -- set frame format to MarkIV format 
                                 (default : depend on info file)
               -long         -- set 64 bit(track) mode
                                 (default 32 bit(track) mode)

               -date jjj|yyyymmdd|yyyyddd  -- 
                     jjj      : lower 3digits of MJD 
                       or
                     yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month
                       or
                     yyyyddd  : 4 digit year, 3 digit day of year


                -monit        -- info_file monitor ON 


               *** following command is not implemented yet **
               -k1 k5name    -- K5 data file name containing Ch01-04
               -k2 k5name    -- K5 data file name containing Ch05-08
               -k3 k5name    -- K5 data file name containing Ch09-12
               -k4 k5name    -- K5 data file name containing Ch13-16
                   or
               -k  k5name    -- K5 data file name (upto 4 file names)

               *********

                VLBA mode :  parity ON,  NRZM   20160/frame
                MarkIV    :  parity off, NRZL   20000/frame


  Enviroment variables
          M5DIR    ---   default path to Mark5 data
          M5VEX    ---   default path to Mark5 vex file


  AD bit is limited to 2bit max. 

  === Default Conversion rule ===

  TRACK# = 32

  AD = 1bit mode 
  bitpos     0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
  K5 CH#   1-1 1-2 1-3 1-4 2-1 2-2 2-3 2-4 3-1 3-2 3-3 3-4 4-1 4-2 4-3 4-4
  bitpos    16  17  18  19  20  21  22  23  24  25  26  27  28  29  30  31
  K5 CH#   --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
    
  AD = 2bit mode 
  bitpos     0    8    2   10    4   12    6   14 
  K5 CH#  1-1s 1-1m 1-2s 1-2m 1-3s 1-3m 1-4s 1-4m
  bitpos     1    9    3   11    5   13    7   15 
  K5 CH#  2-1s 2-1m 2-2s 2-2m 2-3s 2-3m 2-4s 2-4m
  bitpos    16   24   18   26   20   28   22   30 
  K5 CH#  3-1s 3-1m 3-2s 3-2m 3-3s 3-3m 3-4s 3-4m
  bitpos    17   25   19   27   21   29   23   31 
  K5 CH#  4-1s 4-1m 4-2s 4-2m 4-3s 4-3m 4-4s 4-4m

  where  CH# :  A-B  or A-BC   
                where  A: K5 PC ID (#)
                       B: K5 CH# (1-4)
                       C: s = sign bit
                          m = magnitude bit


*** sample run parameters ****

Example 1   date set by 3-digit MJD method
927 O2880001.k5a O2880001.k5b O2880001.k5c O2880001.k5d m5test.dat2

Example 2   date set by YYYYDDD method
2003288 O2880001.k5a O2880001.k5b O2880001.k5c O2880001.k5d m5test.dat2

Example 3   date set by YYYYMMDD method
20031015 O2880001.k5a O2880001.k5b O2880001.k5c O2880001.k5d m5test.dat2


20041113 G:\mark5\k5-2bit.dat -o G:\mark5\m5test.dat2 -mark4
20041113 G:\mark5\k5-2bit.dat -o G:\mark5\m5test.dat2 -i G:\mark5\k5tom5info.txt
-i -G:\mark5\k5tom5info.txt -v G:\mark5\gg057c.vex -sid Ks
20041113 G:\mark5\k5-2bit.dat2.1 G:\mark5\k5-2bit.dat2.2 G:\mark5\k5-2bit.dat2.3 G:\mark5\k5-2bit.dat2.4 -o G:\mark5\m5test.dat3 -i G:\mark5\k5tom5info.txt
20041113 G:\mark5\k5-2bit.dat2.1 G:\mark5\k5-2bit.dat2.2 G:\mark5\k5-2bit.dat2.3 G:\mark5\k5-2bit.dat2.4 -o G:\mark5\m5test.dat3r -i G:\mark5\k5tom5Hinfo.txt -monit
20041113 G:\mark5\k5-2bit.dat -o G:\mark5\m5test.dat2 -i G:\mark5\k5tom5info.txt
*/

#define PROGRAM_VERSION "k5tom5 (Ver 1.21  2005-09-17)"

#include "libk5.h"      // common header for IP-VLBI utilities
#include "libmk5tok5.h"  // library for Mark5 to K5 conversion

/* default conversion information file name set */
#define INFONAME_DFLT "k5tom5info.txt"   // information file default name
/* environment variable for Mark5 data default directory set */
#define M5DIR "M5DIR" // the environment variable for Mark5 data default directory
/* environment variable for Mark5 vex file default directory set */
#define M5VEX "M5VEX" // the environment variable for Mark5 vex file
/* default pass set */
#define DEFLT_PASS "A"   /* default pass in $TRACKS in a VEX file (added on 2004-06-29) */

/* default parameters  */
#define K5NUMCH  4
#define M5PARITY 0
#define M5VLBA   0
#define M5FANOUT 1
#define M5ADBIT  1
#define M5SAMPLE 4   // default sampling frequency in MHz
#define M5NTRACK 32

//#define HUYGENS  // Huygens processing mode

void param_get_k5tom5(int argc, char *argv[],k5m5runp_t *runp)                       
{
    int n,i;
    int len;
    int mode;
    char cbuf[60],buf[10];
    double dw; // work

    // initialize parameters

    /* common parameters */
    runp->ptype=1;  // run with information file
    for(n=0; n<4; n++) runp->k5name[n][0]='\0';
    runp->mk5name[0]='\0';

    runp->soffset=0;
    runp->period=0;
    runp->odd=0;

    /* mode 1 parameters */
    runp->channel=0;
    runp->group=0;
    runp->reverse=0;
    runp->makemode=0;
    runp->info_file[0]='\0';
    runp->outdir[0]='\0';
    runp->vex_file[0]='\0';
    runp->sid[0]='\0';
    runp->scan=1;
    runp->monit=0;
    strcpy(runp->dflt_pass, DEFLT_PASS);

    /* mode 2 parameters */
    runp->numch=K5NUMCH;
    runp->ntrack=M5NTRACK;
    runp->adbit =M5ADBIT;
    runp->parity=M5PARITY;
    runp->nrzm  =-1;
    runp->fanout=M5FANOUT;
    runp->vlba  =M5VLBA;
    runp->tot_track=0;
    runp->fmhz  =(double)M5SAMPLE;
    for(n=0; n<32; n++) runp->ibit[n]=0;

    runp->yyyy=0; runp->mm=0; runp->dd=0; runp->ddd=0;
    runp->mjd=0; runp->mjd3=0;

    runp->nbits=32;

    mode=0;
    /* 1st param check if it is started with "-" or not  */
    if (argc>=2){
        strcpy(cbuf,*(argv+1));
        if(cbuf[0] == '-') mode=1;
    }

    if(mode == 1) {
        n=2;
        goto comm_anal;
    }

    /* 1st param should be either jjj(3digit MJD) or yyyymmdd or yyyyddd */
    if (argc>=2){
        strcpy(cbuf,*(argv+1));
        len=strlen(cbuf);
        //printf("len = %d\n",len);
        if(len==8){ // yyyymmdd
            strncpy(buf,cbuf,4); buf[4]='\0';
            runp->yyyy=atoi(buf);
            strncpy(buf,cbuf+4,2); buf[2]='\0';
            runp->mm=atoi(buf);
            strncpy(buf,cbuf+6,2); buf[2]='\0';
            runp->dd=atoi(buf);
            //get MJD
            runp->mjd=ut2mjd(&dw,runp->yyyy,runp->mm,runp->dd,0,0,0);
            //get total day 
            md2ddd(runp->yyyy,&runp->ddd,runp->mm,runp->dd);
        } else if(len==7){ // yyyyddd
            strncpy(buf,cbuf,4); buf[4]='\0';
            runp->yyyy=atoi(buf);
            strncpy(buf,cbuf+4,3); buf[3]='\0';
            runp->ddd=atoi(buf);
            //get month day
            ddd2md(runp->yyyy,runp->ddd,&runp->mm,&runp->dd);
            //get MJD
            runp->mjd=ut2mjd(&dw,runp->yyyy,runp->mm,runp->dd,0,0,0);
        } else {
            runp->mjd = atoi(cbuf);
        }
        runp->mjd3 = runp->mjd % 1000;
    }

    n=3;
    runp->k5num=0;

    /* K5 name get */
    while(1){
        if(argc >= n) {
           strcpy(cbuf,*(argv+ n-1) );
           //printf("%s\n",cbuf);
           if(cbuf[0] == '-') break;  // go to command parameter get
           //if(runp->k5num >= 4) {
           //    printf(" Warning!! Number of K5 file exceeds 4!!\n");
           //    printf(" So more K5 files will be neglected!!\n");
           //    runp->k5num=4;
           //    n++;
           //    break;
           //}
           if(runp->k5num >= 4) {  // mark5 name
               strcpy(runp->mk5name,cbuf);
               runp->k5num=4;
               n++;
               break;
           }
           strcpy(runp->k5name[runp->k5num],cbuf);
           runp->k5num++;
        } else {
            goto exitend;
        }
        n++;
    }


comm_anal:


    /* command parameters get */
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        if(strncmp(cbuf,"-i",2)==0){  // info file
            n++;
            if(n > argc) break;
            strcpy(runp->info_file,*(argv+n-1));
            if(runp->info_file[0]=='-'){
                runp->makemode=1;
                len=strlen(runp->info_file);
                for(i=0;i<len-1;i++) runp->info_file[i]=runp->info_file[i+1];
                runp->info_file[len-1]='\0';
            } else if(strncmp(runp->info_file,"make",4)==0){
                runp->makemode=1;
                runp->info_file[0]='\0';
            }
        }
        if(strlen(cbuf)==2 && strncmp(cbuf,"-o",2)==0){  // Mark5 name
            n++;
            if(n > argc) break;
            strcpy(runp->mk5name,*(argv+n-1));
             /* check Mark5 file name is acompanied by directory or not */
            get_true_path(runp->mk5name, M5DIR, "./",cbuf);
            strcpy(runp->mk5name,cbuf);
        }
        if(strncmp(cbuf,"-d",2)==0){  // out directory
            n++;
            if(n > argc) break;
            strcpy(runp->outdir,*(argv+n-1));
            /* env variable is substituted by real one */

            if((i=substitute_env_path_name(runp->outdir, cbuf))>0){
                strcpy(runp->outdir,cbuf);
            }

        }
        if(strncmp(cbuf,"-c",2)==0){  // channel define
            n++;
            if(n > argc) break;
            runp->channel = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-g",2)==0){  // group define
            n++;
            if(n > argc) break;
            runp->group = atoi(*(argv + n-1 ));
        }
        if(strlen(cbuf)==2 && strncmp(cbuf,"-s",2)==0){  // start offset
            n++;
            if(n > argc) break;
            runp->soffset = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-p",2)==0){  // period
            n++;
            if(n > argc) break;
            runp->period = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-r",2)==0){  // 
            runp->reverse=1;
        }
        if(strncmp(cbuf,"-v",2)==0){  // vex file
            n++;
            if(n > argc) break;
            strcpy(runp->vex_file,*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-odd",4)==0){  // odd mode set
            runp->odd=1;
        }
        if(strncmp(cbuf,"-parity",3)==0){  // parity
           runp->parity=1;
        }
        if(strncmp(cbuf,"-noparity",3)==0){  // non-parity
           runp->parity=0;
        }
        if(strncmp(cbuf,"-nrzm",5)==0){  // NRZM
           runp->nrzm=1;
        }
        if(strncmp(cbuf,"-nrzl",5)==0){  // NRZL
           runp->nrzm=0;
        }
        if(strncmp(cbuf,"-vlba",3)==0){  // VLBA mode
           runp->vlba=1;
        }
        if(strncmp(cbuf,"-mark4",6)==0){  // Mark4 mode
           runp->vlba=0;
        } 
        if(strncmp(cbuf,"-m4",3)==0){  // Mark4 mode
           runp->vlba=0;
        } 
        if(strncmp(cbuf,"-monit",2)==0){  // monitor mode on
            runp->monit=1;
        }
        if(strncmp(cbuf,"-long",3)==0){  // 64bit(track) mode on
            runp->nbits=64;
        }
        if(strncmp(cbuf,"-si",3)==0){  // station id
            n++;
            if(n > argc) break;
            strcpy(runp->sid,*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-scan",5)==0){  // scan#
            n++;
            if(n > argc) break;
            runp->scan=atoi(*(argv + n-1 ));
            if(runp->scan < 1) runp->scan=1;
        }
        n++;
    }

exitend:;

}

void print_usage(void)
{
    printf("%s\n",PROGRAM_VERSION);
    printf("\n");
    printf("How to use\n");
    printf("\n");
    //printf("    k5tom5 yymmdd k51name [k52name k53name k54name] [options]\n");
    printf("    k5tom5 jjj|yyyymmdd|yyyyddd\n");
    printf("                   k5name1 [k5name2 k5name3 k5name4 [m5name]] [options]\n");
    printf("        or\n");
    printf("    k5tom5 -i -infofile|make -v vexfile [options]\n");
    printf("               .... for information file creation mode\n");
    printf("\n");
    //printf("where  yymmdd ---- year(2digits),month(2digits),day of month (2digits) of\n");
    //printf("                      data start time\n");
    printf("   where\n");
    printf("       jjj|yyyymmdd|yyyyddd  -- observation date in three ways as follows\n");
    printf("           jjj      : lower 3digits of MJD \n");
    printf("              or\n");
    printf("           yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month\n");
    printf("              or\n");
    printf("           yyyyddd  : 4 digit year, 3 digit day of year\n");
    printf("       k5name1 ---- K5 file name for CH01-04 or CH01\n");
    printf("       k5name2 ---- K5 file name for CH05-08 or CH02\n");
    printf("       k5name3 ---- K5 file name for CH09-12 or CH03\n");
    printf("       k5name4 ---- K5 file name for CH13-16 or CH04\n\n");
    printf("       m5name  ---- Mark5 file name to be created\n\n");
    printf("      options (any order)\n");

    printf("        -i info_file  -- infomation file name that contains\n");
    printf("                           track vs channel table\n");
    printf("                           bit_psition vs channel table\n");
    printf("                           group# vs channel table\n");
    printf("                           (default \"%s\")\n",INFONAME_DFLT);
    printf("                           \"-i -info_file\" means\n");
    printf("                             create infomation file named info_file\n");
    printf("                           \"-i make\" means create default info file\n");

    printf("        -o m5name     -- Mark5 file name to be created\n");
    //printf("                          (default : see below)\n");
    printf("        -d m5dir      -- directory for Mark5 file out\n");
    printf("                          (default  : mark5 directory)\n");
    printf("        -s strat_sec  -- offset time from data head (sec)\n");
    printf("                          (default :  0 )\n");
    printf("        -p period     -- data period to convert (sec)\n");
    printf("                          (default : all)\n");
    printf("        -mark4        -- set frame format to Mark4 mode (default VLBA mode)\n");
    printf("        -long         -- set 64 bit(track) mode (default 32bit mode)\n");
    printf("        -v vex_file   -- VEX file name to be read when \"information\n");
    printf("                              file creation mode\" is set\n");
    printf("        -sid stat_id  -- station ID when \"information file creation\n");
    printf("                              mode\" is set\n");
    printf("        -scan scan#   -- scan# for MODE get when \"information file\n");
    printf("                              creation mode\" is set (default : 1)\n");
    //printf("        -monit        -- info_file monitor ON\n");
    printf("\n");
}

void monit_runp(k5m5runp_t *runp)
// run parameter monitor
{
    int n;

       printf(" ---------- RUN CONDITION ----------\n");
       printf("   yyyymmdd  : %04d%02d%02d\n",runp->yyyy,runp->mm,runp->dd);
       printf("   yyyyddd   : %04d%03d\n",runp->yyyy,runp->ddd);
       printf("   jjj       : %03d\n",runp->mjd3);
       if(runp->yyyy != 0){
       printf("   MJD       : %d\n",runp->mjd);
       }
       for(n=0; n<runp->k5num; n++)
       printf("   k5name %d  : %s\n",n+1,runp->k5name[n]);

       printf("   mk5name   : %s\n",runp->mk5name);
       printf("   outdir    : %s\n",runp->outdir);
       printf("   infofile  : %s\n",runp->info_file);
       printf("   channel   : %d\n",runp->channel);
       printf("   group     : %d\n",runp->group);
       printf("   soffset   : %d\n",runp->soffset);
       printf("   period    : %d\n",runp->period);
       printf("   odd       : %d\n",runp->odd);
       printf("   parity    : %d\n",runp->parity);
       printf("   nrzm      : %d\n",runp->nrzm);
       printf("   vlba      : %d\n",runp->vlba);
       printf("   nbits     : %d\n",runp->nbits);
       if(runp->vex_file[0] != '\0')
          printf("   vex_file  : %s\n",runp->vex_file);
       if(runp->sid[0] != '\0' && runp->sid[0] !='-')
          printf("   station_id: %s\n",runp->sid);
       printf(" -------------------------------------\n");

    printf("\n");
}

void monit_head(m5head_t *hd)
{
    //monitor Mark5 header
    int i,k,n,m;
    int head[5],sync[36];
    int header[180];

    // decode NRZM
    header[0]=0;
    for(i=1; i<180; i++){
        header[i]= hd->ihead[i-1] ^ hd->ihead[i];
    }

    for(i=0; i<5; i++){
       for(k=0; k<36; k++){
          sync[k]=  header[i*36+k] & 0x1;  // pick up LSB
       }
       head[i]=conv2int(sync,1);
    }

    printf("                                  TIME CODE  CRC-  \n");
    printf("      HEADER     AUX     SYNC   MJDsssss.ssss  16  DATA#1   DATA#2   FMHz\n");
    printf("     ");
    for (i=0; i<5; i++){   //check           
       printf("%08X ",rvsint(head[i]));
    }   
    printf("\n");

    // monitor by every bit

    printf("\n");
    printf("  #  01234567 P\n");
    printf(" --  -------- -\n");
    m=0;
    for(i=0; i<5; i++){
        for(n=0; n<4; n++){
            printf(" %2d  ",m);
            m++;
            for(k=0; k<9; k++){
                printf("%d",header[i*36+n*9+k] & 0x1);
                if(k==7) printf(" ");
            }
            printf("\n");
       }
    }
    printf(" --  -------- -\n");
}

void parseK5(unsigned char *ia, int numb, int adbit, int numch, int *xdata, int *kpos)
{
/*
    Parese K5 data to bit stream corresponding to track stream data in Mark5

      [in]
         ia  ---  sampler K5 out data 
         numb --  size of ia (bytes)
         adbit -- sampler A/D resolution bits (1,2,4,8)
      [return]
         xdata -- data parsed into each bit stream
         kpos  -- valid sample number of data in xdata array 

     Relation between # of ch (numch) and each ch time siries

        # of ch = numch, # of AD bits = adbit, then
        ch# m, AD bit# n, k-th data is given

        sdata(m,n,k)=xdata(m+n*numch+k*numch*adbit)

        
*/
     int inumb, pat, n, i, ww, k, m;
     unsigned int *ibuf;

     inumb=numb/4;

     ibuf=(unsigned int *)ia;


     k=0;
     for( i=0; i < inumb; i++){
        ww=ibuf[i];
        pat=0;
        while(pat<32){
           for(m=0; m<numch; m++){
              for(n=0; n<adbit; n++){
                 xdata[m+n*numch+k*numch*adbit]=ww & 0x1;
                 pat++;
                 ww>>=1;
              }
           }
           k++;
        }
     }

     *kpos=k;
     //printf("numb8,ipos %d %d\n",numb8,ipos);
}

void mon_parseK5(int *xdata, int adbit, int numch)
{
/*
    Monitor of paresed K5 data to bit stream corresponding to track in Mark5
*/
    int a,k,m,sdat;
    for(m=0; m<numch; m++){
        for(a=0; a<adbit; a++){
           printf("CH# %d   AD bit# %d\n",m+1,a);
           for(k=0; k<100; k++){ 
              sdat=xdata[m+a*numch+k*numch*adbit];
              printf(" %d",sdat);
              if((k+1)%20 ==0) printf("\n");
           }
        }
    }
}

int get_bitpos(int m5ch, int adinfo, int fo_pos, int *track, int *hstack)
{
/*
      m5ch  -- Mark5 channel  1,2,....
      adinfo --  0: sign bit   1: magnitude bit
      fo_pos --  fan out position   1,2,3,4

    return
      track  -- track#  2,..,33
      hstack -- head stack number 1,2

      get_bitpos -- bit position 


  tarck table for   AD 2bit fanout=4
     fanout_def = A : &CH01 : sign : 1:  2:  4:  6:  8;
     fanout_def = A : &CH01 :  mag : 1: 10: 12: 14: 16;
     fanout_def = A : &CH02 : sign : 1: 18: 20: 22: 24;
     fanout_def = A : &CH02 :  mag : 1: 26: 28: 30: 32;
     fanout_def = A : &CH03 : sign : 1:  3:  5:  7:  9;
     fanout_def = A : &CH03 :  mag : 1: 11: 13: 15: 17;
     fanout_def = A : &CH04 : sign : 1: 19: 21: 23: 25;
     fanout_def = A : &CH04 :  mag : 1: 27: 29: 31: 33;
     fanout_def = A : &CH05 : sign : 2:  2:  4:  6:  8;
     fanout_def = A : &CH05 :  mag : 2: 10: 12: 14: 16;
     fanout_def = A : &CH06 : sign : 2: 18: 20: 22: 24;
     fanout_def = A : &CH06 :  mag : 2: 26: 28: 30: 32;
     fanout_def = A : &CH07 : sign : 2:  3:  5:  7:  9;
     fanout_def = A : &CH07 :  mag : 2: 11: 13: 15: 17;
     fanout_def = A : &CH08 : sign : 2: 19: 21: 23: 25;
     fanout_def = A : &CH08 :  mag : 2: 27: 29: 31: 33;
*/
       int trk_table[]={  2, 4, 6, 8,
                         10,12,14,16,
                         18,20,22,24,
                         26,28,30,32,
                          3, 5, 7, 9,
                         11,13,15,17,
                         19,21,23,25,
                         27,29,31,33};
       int ch,k,bitpos;
       *hstack=(int)((m5ch-1)/4)+1;
       ch= (m5ch-1) % 4;    // CH# 0,1,2,3
       k=ch*8+4*adinfo+(fo_pos-1);
       *track=trk_table[k];
       bitpos=(*hstack-1)*32+(*track-2);
       return bitpos;
}


int get_bitpos2(int m5ch, int adinfo, int fo_pos, mk5tok5_info_t *mk5info,
                m5head_t *hd)
{
/*
      m5ch  -- Mark5 channel  1,2,....
      adinfo --  0: sign bit   1: magnitude bit
      fo_pos --  fan out position   1,2,3,4
      mk5info -- mk5info
      hd      -- mark5 header info

    return
      track  -- track#  2,..,33
      hstack -- head stack number 1,2

      get_bitpos2 -- bit position 

*/
       int n,bitpos;
       bitpos=0;
       for(n=0; n<hd->nbits; n++){
           if(hd->chinfo[n]==m5ch &&
               hd->adinfo[n]==adinfo &&
               hd->faninfo[n]==fo_pos){
               bitpos=n;
               break;
           }
       }
       return bitpos;
}


void head_check(m5head_t *hd)
{
    int a,k,kk,m,n;
    int hdchk[5*64];
    int set_table[]={ 0x00000001,0x00000002,0x00000004,0x00000008,
                         0x00000010,0x00000020,0x00000040,0x00000080,
                         0x00000100,0x00000200,0x00000400,0x00000800,
                         0x00001000,0x00002000,0x00004000,0x00008000,
                         0x00010000,0x00020000,0x00040000,0x00080000,
                         0x00100000,0x00200000,0x00400000,0x00800000,
                         0x01000000,0x02000000,0x04000000,0x08000000,
                         0x10000000,0x20000000,0x40000000,0x80000000};


    //below is for check
    if(hd->vlba==1) goto exithere; // nothing to do
    for(n=0; n<320; n++) hdchk[n]=0;
    for(n=0; n<160; n++){
        k=n/32;
        a=31-(n % 32);
        //a=(n % 32);
        for(m=0; m<hd->nbits; m++){
            if(hd->nbits==64){
                kk= (int)(hd->lhead[n] >> m) & 1;
            } else {
                kk= (int)(hd->ihead[n] >> m) & 1;
            }
            if(kk == 1){
                hdchk[m*5+k]  |= set_table[a];
            }
        }
    }
    for(m=0; m<hd->nbits; m++){
        printf(" %2d",m);
        for(k=0; k<5; k++){
            //printf(" %08X",rvsint(hdchk[m*5+k]));
            printf(" %08X",hdchk[m*5+k]);
        }
        printf("\n");
    }
exithere:;
}


void set_hd_trk_info(mk5tok5_info_t *mk5info, m5head_t *hd)
{
   //  setup mark5 headder track info
   int n,k,kk,m,trk,hs;

   for(n=0; n<hd->nbits; n++){
       hd->trkinfo[n]=mk5info->bit_track[n];
       hd->headstackinfo[n]= mk5info->bit_hs[n];

       hd->adinfo[n]=0; // sign bit (default)
       hd->chinfo[n]=0; // CH# 
       hd->vcinfo[n]=0; // BBC#
       hd->lsbinfo[n]=0; // side band info

       trk=mk5info->bit_track[n];   // track#
       hs =mk5info->bit_hs[n];       // head stack#
       if(trk > 0){
           for(k=0; k<16; k++){
               kk=0;
               for(m=0; m<hd->fanout; m++){
                   if(trk==mk5info->signtrack[k][m] &&
                       hs==mk5info->hs_sign[k][m]){
                       hd->faninfo[n]=m+1;  // fanout info
                       kk=1;
                       break;
                   }
                   if(trk==mk5info->magtrack[k][m] &&
                       hs==mk5info->hs_mag[k][m]){
                       hd->faninfo[n]=m+1;  // fanout info
                       hd->adinfo[n]=1;  // mag bit
                       kk=1;
                       break;
                   }
               }
               if(kk==1){
                   //ch=k+1; //CH#
                   //hd.chinfo[n]=ch;
                   hd->chinfo[n]=mk5info->ch_ch[k];
                   hd->vcinfo[n]=mk5info->ch_bbc[k];
                   hd->lsbinfo[n]=mk5info->ch_lsb[k];
                   break;
               }
           }
       }
   }
}

void setup_work_table(int fanout, int adbit, int *bittable,
                     mk5tok5_info_t *mk5info, m5head_t *hd)
{
    /*work table set up */
    int i,n,fo_pos,a,m5ch,adinfo,bitpos;
    i=0;
    for(fo_pos=1; fo_pos<=fanout; fo_pos++){
        //for(n=0; n<runp.k5num; n++){
        //for(n=0; n<16; n++){
        for(n=0; n<mk5info->nch; n++){
            //m5ch=n+1;  // Mark5 CH#  (1,2,....)
            m5ch=mk5info->ch_ch[n];
            //printf("m5ch= %d\n",m5ch);
            for(a=0; a<adbit; a++){   // adbit=1   a=0: sign bit
                                      // adbit=2   a=0: mag bit
                                      //           a=1: sign bit
                adinfo=adbit-a-1; // adinfo  =0: sign bit  =1: mag bit
                //bitpos=get_bitpos(m5ch,adinfo,fo_pos,&track,&hstack);
                bitpos=get_bitpos2(m5ch,adinfo,fo_pos,mk5info,hd);
                bittable[(fo_pos-1)+(m5ch-1)*fanout+adinfo*fanout*16]=bitpos;
                i++;
            }
        }
    }
    //printf("bittable size %d\n",i);  // debug
}

 
void put_k5_data(int adbit,  int m5ch, int fanout, int fo_pos, int k, int m,
                   int *pdat, int *bittable,
                   int *iw, int *set_table,
                   int64_t *lw, int64_t *set_ltable, ip_vlbi_file_old_t *k5,
                   m5head_t *hd)
{
    int a, adinfo, sdat, bitpos;

    for(a=0; a<adbit; a++){   // adbit=1   a=0: sign bit
                              // adbit=2   a=0: mag bit
                              //           a=1: sign bit
         adinfo=adbit-a-1; // adinfo  =0: sign bit  =1: mag bit
         sdat=k5->idata[m+a*k5->numch+k*k5->numch*adbit];
         if(sdat==1){ //set data at bitpos
            //bitpos get
            //bitpos=get_bitpos(m5ch,adinfo,fo_pos,&track,&hstack);
            bitpos=bittable[(fo_pos-1)+(m5ch-1)*fanout+adinfo*fanout*16];
            //if(bitpos >=64) printf(" bitpos %d",bitpos);//debug
            if(hd->nbits==64){
               *lw |= set_ltable[bitpos];
            } else {
               *iw |= set_table[bitpos];
            }
            pdat[bitpos]++;  // count # of 1 for parity
         }
    }
}
 
void put_sine_wave(double t, int adbit,  int m5ch, int fanout, int fo_pos,
                   int *pdat, int *bittable,
                   int *iw, int *set_table,
                   int64_t *lw, int64_t *set_ltable, m5head_t *hd)
{
    int a, adinfo, iy, sdat, bitpos;
    //put 1MHz dummy sine signal
    //iy=(int)(1.5*sin(fmod(TWOPI*1.0e6*t,TWOPI))+2.0); // for adbit=2
    //iy=(int)(0.5*sin(fmod(TWOPI*1.0e6*t,TWOPI))+1.0); // for adbit=1
    iy=(int)((0.5+(double)(adbit-1))*sin(fmod(TWOPI*1.0e6*t,TWOPI))+1.0+(double)(adbit-1)); 
    for(a=0; a<adbit; a++){   // adbit=1   a=0: sign bit
                              // adbit=2   a=0: mag bit
                              //           a=1: sign bit
        adinfo=adbit-a-1; // adinfo  =0: sign bit  =1: mag bit
        if(a==0) { //lsb
           sdat=iy % 2;
        } else {
           sdat=iy/2;
        }
        if(sdat==1){ //set data at bitpos
           //bitpos get
           //bitpos=get_bitpos(m5ch,adinfo,fo_pos,&track,&hstack);
           bitpos=bittable[(fo_pos-1)+(m5ch-1)*fanout+adinfo*fanout*16];
           //if(bitpos >=64) printf(" bitpos %d",bitpos);//debug
           if(hd->nbits==64){
              *lw |= set_ltable[bitpos];
           } else {
              *iw |= set_table[bitpos];
           }
           pdat[bitpos]++;  // count # of 1 for parity
        } 
    }
}
 
int main(int argc, char *argv[])
{
       int i, k, kk, n,m,a;
       int ch;
       int aflag;
       int secmin,secmax;
       int iread,iwrite;
       int kpos;
       int sdat,bitpos;
       int pdat[64]; // save parity calc data
       int pk; // for parity calc
       unsigned int iw;
       int64_t lw;
      //unsigned int obuf[9];
       //unsigned int obuf[2250]; // 2250=2000*9/8
       unsigned int obuf[225]; // 225=200*9/8   This shows max performance
       int obufmax=225;
       int ow;
       int64_t lbuf[225];

       int period;
       int fanout;

       int sfreq, adbit, numch;
       int imax, isa;
       int usampl;
       long int samplpp;  // samples/PP
       int bits, bytes, numb;
       int tmax,tmin;
       int m5ch,adinfo,fo_pos,koffset,owp;
       //int bittable[64];
       int bittable[4*2*16];  // fanout*adbit*maxCH
       int filen_tabl[16];  // K5 file number index table for ch processing order
       int k5chn_tabl[16];  // K5 ch# in a file index table for ch processing order
       double fsampl,tsampl,unitp;
       double t;
       
       double telapsed;           // to measure processing time
       char fname[20];
       char fpath[100];
       char program_version[60];
       char* dflt_iname=INFONAME_DFLT;   // information file default name
	   time_t tstart, tfinish;    
       mk5tok5_info_t mk5info;
       //gr_info_t gr; 
       k5m5runp_t runp;
       FILE *fo;
       ip_vlbi_file_old_t k5[4];   // IP-VLBI data type variables 
       m5head_t hd;
       int set_table[]={ 0x00000001,0x00000002,0x00000004,0x00000008,
                         0x00000010,0x00000020,0x00000040,0x00000080,
                         0x00000100,0x00000200,0x00000400,0x00000800,
                         0x00001000,0x00002000,0x00004000,0x00008000,
                         0x00010000,0x00020000,0x00040000,0x00080000,
                         0x00100000,0x00200000,0x00400000,0x00800000,
                         0x01000000,0x02000000,0x04000000,0x08000000,
                         0x10000000,0x20000000,0x40000000,0x80000000};
#ifdef WIN32
       int64_t set_ltable[]={ 
        0x1,0x2,0x4,0x8,
        0x10,0x20,0x40,0x80,
        0x100,0x200,0x400,0x800,
        0x1000,0x2000,0x4000,0x8000,
        0x10000,0x20000,0x40000,0x80000,
        0x100000,0x200000,0x400000,0x800000,
        0x1000000,0x2000000,0x4000000,0x8000000,
        0x10000000,0x20000000,0x40000000,0x80000000,
        0x100000000,0x200000000,0x400000000,0x800000000,
        0x1000000000,0x2000000000,0x4000000000,0x8000000000,
        0x10000000000,0x20000000000,0x40000000000,0x80000000000,
        0x100000000000,0x200000000000,0x400000000000,0x800000000000,
        0x1000000000000,0x2000000000000,0x4000000000000,0x8000000000000,
        0x10000000000000,0x20000000000000,0x40000000000000,0x80000000000000,
        0x100000000000000,0x200000000000000,0x400000000000000,0x800000000000000,
        0x1000000000000000,0x2000000000000000,0x4000000000000000,0x8000000000000000};
#else 
       int64_t set_ltable[]={ 
        0x1LL,0x2LL,0x4LL,0x8LL,
        0x10LL,0x20LL,0x40LL,0x80LL,
        0x100LL,0x200LL,0x400LL,0x800LL,
        0x1000LL,0x2000LL,0x4000LL,0x8000LL,
        0x10000LL,0x20000LL,0x40000LL,0x80000LL,
        0x100000LL,0x200000LL,0x400000LL,0x800000LL,
        0x1000000LL,0x2000000LL,0x4000000LL,0x8000000LL,
        0x10000000LL,0x20000000LL,0x40000000LL,0x80000000LL,
        0x100000000LL,0x200000000LL,0x400000000LL,0x800000000LL,
        0x1000000000LL,0x2000000000LL,0x4000000000LL,0x8000000000LL,
        0x10000000000LL,0x20000000000LL,0x40000000000LL,0x80000000000LL,
        0x100000000000LL,0x200000000000LL,0x400000000000LL,0x800000000000LL,
        0x1000000000000LL,0x2000000000000LL,0x4000000000000LL,0x8000000000000LL,
        0x10000000000000LL,0x20000000000000LL,0x40000000000000LL,0x80000000000000LL,
        0x100000000000000LL,0x200000000000000LL,0x400000000000000LL,0x800000000000000LL,
        0x1000000000000000LL,0x2000000000000000LL,0x4000000000000000LL,0x8000000000000000LL};
#endif

        int m5ch_ok_tabl[]={ 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
                           // if you want to omit some Mark5 channel
                           // you can edit this table directly (set 0)

        if(argc == 1){
          print_usage();
          exit(0);
        } 
       if (argc>=2){
          if(strncmp(*(argv+1),"env",3)==0){
          printf("   Enviroment variables\n");
          printf("        %s ---   default directory for Mark5 data\n",M5DIR);
          printf("                      ( %s )\n",getenv(M5DIR));
          printf("        %s ---   default directory for Mark5 vex file\n",M5VEX);
          printf("                      ( %s )\n",getenv(M5VEX));
          printf("\n");
          exit(0);
          }
       }

       strcpy(program_version,PROGRAM_VERSION);

       /* run parameters get */
       param_get_k5tom5(argc, argv, &runp); 

       if (runp.info_file[0]=='\0'){
          strcpy(runp.info_file,dflt_iname);
       }

       /* out directory create if not set */

       if(runp.mk5name[0] != 0 && runp.outdir[0]=='\0'){
          parse_dir_and_name(runp.mk5name, runp.outdir, fname); // debug
          printf("A %s\n",runp.mk5name);
          printf("A %s\n",runp.outdir);
          printf("A %s\n",fname);
       }

       /* create an information file if selected so */
       if(runp.makemode ==1){  //
           /*****************************************************/
           /* Entry of Information file creation mode operation */ 
           /*****************************************************/
           printf("K5tom5 running under Information File create mode\n");
           printf("info file (%s) will be created (updated).\n",runp.info_file);
           if(runp.vex_file[0]=='\0'){  // interactive operation
              printf("\nEnter VEX file name -> ");
              scanf("%s",&runp.vex_file);
           } 

          /* get true path for vex file name  */
           get_true_path(runp.vex_file, M5VEX, "./",fpath);
           strcpy(runp.vex_file,fpath);
           printf("VEX file name --- %s\n",runp.vex_file);

           make_k5tom5_info_file(runp.info_file,runp.vex_file,
               runp.sid,runp.scan,program_version,runp.dflt_pass);
           printf("\nInformation file (%s) created!!\n",runp.info_file);
           exit(1);
           /***************************************************/
           /* End of Information file creation mode operation */ 
           /***************************************************/
       }

       /**************************************/
       /* Entry of Conversion mode operation */ 
       /**************************************/


       if(runp.ptype == 1){ // in case of mode 1 
           /* internal conversion information set up */
           k5tom5_info_setup(runp.info_file, &mk5info);

          /* monitor mk5info */
          if(runp.monit==1)
              monit_mk5info(&mk5info);  // debug    
       } else {    // in case of mode 2 
           /*
           strcpy(runp.k5name,runp.k5name_given);
           */
           /* set parameters for mode 2 operation */ 
           /*
           if(runp.nrzm==-1){ // automatic set this parameter
               if(runp.parity==1) runp.nrzm=1;
               if(runp.parity==0) runp.nrzm=0;
           }
           mk5info.parity=runp.parity;
           mk5info.nrzm=runp.nrzm;  
           mk5info.adbit=runp.adbit;
           mk5info.ntrack=runp.ntrack;
           mk5info.samplrate=runp.fmhz*1.0e6;    // sampling frequency (Hz)
           mk5info.fanout=runp.fanout;
           mk5info.vlba=runp.vlba;
           if(mk5info.vlba==1){
               mk5info.bitsframe=20160;
           } else {
               mk5info.bitsframe=20000;
           }
           */
       }
       /* Program Name Display */
       printf("************************************************************\n");
       printf("*    K5/VSSP to Mark-V Data Format Converter               *\n");
       printf("*      %s  by  T.KONDO/NICT    *\n",program_version);
       printf("************************************************************\n");
       printf("\n");

       monit_runp(&runp);


       tstart= time(NULL);


	   /* K5 data check to get sampling information */

	   for(n=0; n<runp.k5num; n++){
           /* Data file open */
          //k5[n].ibuf  = (unsigned char *)malloc(numb);             
          k5[n].fname=runp.k5name[n];
          if( (k5[n].stream  = fopen( runp.k5name[n], "rb" )) == NULL ){
              printf("Data File %s   Open Error! \n",runp.k5name[n]);
		      goto exitend;
		  }
         /************************************************/
         /* Data header check to get sampling parameters */
         /************************************************/

         if(chk_K5_header(&k5[n]) < 0){
            goto exitend;
         } 

         //printf("%d\n",k5[n].seconds);  // debug
       }

        /********************************/
        /* set internal-use parameters  */
        /********************************/

        sfreq=k5[0].sfreq;       // sampling frequency (kHz)
        adbit=k5[0].adbit;       // A/D bits
        numch=k5[0].numch;       // # of channel in data (1 or 4)
        fsampl=1000.0*(double)sfreq;   // sampling frequency (Hz)
        tsampl=1.0/fsampl;            // sampling period (sec)
        bits=((long)fsampl)*adbit*numch;  // # of bits in 1 sec data except header
        bytes=bits/8;                     // # of bytes in 1 sec data except header

        samplpp=(long int)fsampl;  // # of samples in a second


        /*****************************************/
        /*  set unit # of samples to process     */
        /*****************************************/

        //hd.fanout=1;  //  fan out in Mark 5 
        //fanout=4;  //  fan out in Mark 5 (for Huygens)
        usampl=20000;   // # of samples in a Mark5 frame
        fanout=mk5info.fanout;
        hd.fanout=fanout;
        hd.vlba=mk5info.vlba;



        imax=(sfreq*1000)/(usampl*fanout);        // # of usampls in a second
        numb=fanout*usampl*adbit*numch/8;  // # of bytes in a usampl

        unitp=(double)usampl*tsampl*(double)fanout;   // period in a usampl (sec)

        for(n=0; n<runp.k5num; n++){
           k5[n].ibuf  = (unsigned char *)malloc(numb);
           k5[n].idata = (int *)malloc(fanout*usampl*adbit*numch*sizeof(int));
        }




       // if start time was different each file, then atamadashi is necessary

       /* check start time */
       aflag=0;   // atamadashi flag
       secmin=k5[0].seconds;
       secmax=k5[0].seconds;
       if(runp.k5num > 0) {
	      for(n=1; n<runp.k5num; n++){
              if(k5[n].seconds < secmin) secmin=k5[n].seconds;
              if(k5[n].seconds > secmax) secmax=k5[n].seconds;
          }
          if(secmin != secmax){
              aflag=1;
              printf(" Warning!! start time different file exists!\n");
          }
       }

       if(aflag==1){ // atamadashi at secmax
           for(n=0; n<runp.k5num; n++){
              isa= secmax-k5[n].seconds;
              printf("atamashi %d\n",isa);
              if(fwd_K5_Nsec(k5[n].ibuf, isa, numb, imax, &k5[n]) <0){
                goto exitend;
              }
           }
       }

       // start offset
       if(runp.soffset > 0){
           for(n=0; n<runp.k5num; n++){
              isa= runp.soffset;
              printf("File %d forward %d sec    ",n+1,isa);
              if(fwd_K5_Nsec(k5[n].ibuf, isa, numb, imax, &k5[n]) <0){
                goto exitend;
              }
           }
           printf("\n");
       }

       /* Mark5 file create */

       if(runp.mk5name[0]=='\0'){
           printf(" Mark 5 file not created!\n");
       } else {
          if( (fo  = fopen( runp.mk5name, "wb" )) == NULL ){
              printf("  File  %s  Open Error! \n",runp.mk5name);
  	          goto exitend;
          }
       }


       /* Mark5 header common info setup */

       hd.ifsecframe=(int)(unitp*10000);
       hd.seconds=k5[0].seconds;
       hd.ifsec=0;
       hd.io=0xFFFFFFFF;  // initial value for NRZM coding
#ifdef WIN32
       hd.lo=0xFFFFFFFFFFFFFFFF;  // initial value for NRZM coding
#else
       hd.lo=0xFFFFFFFFFFFFFFFFLL;  // initial value for NRZM coding
#endif
      //hd.parity=1;
//       if(runp.vlba==1) {
//           hd.vlba=1; 
//           hd.parity=1;
//       } else {
//           hd.vlba=0;
//           hd.parity=0;
//       }
       hd.yyyy=runp.yyyy;
       hd.ddd=runp.ddd;
       hd.mjd=runp.mjd3;

       //hd.nbits=32;  // bits per word.   now supports 32 and 64 bit mode only
       //hd.nbits=runp.nbits;  // bits per word.   now supports 32 and 64 bit mode only
       hd.nbits=mk5info.ntrack;  // bits per word.   now supports 32 and 64 bit mode only

       hd.numch=mk5info.nch;  // # of Mark5 channels


       /* set up track info etc */
       set_hd_trk_info(&mk5info, &hd);
       
/*     //debug
       printf(" bb hs trk ch bbc AD sb fan\n");
       for(n=0; n<hd.nbits; n++){ // debug
           printf("%2d %2d %2d %2d %2d %2d %2d %2d\n",
               n,hd.headstackinfo[n],hd.trkinfo[n],
               hd.chinfo[n],hd.vcinfo[n],hd.adinfo[n],
               hd.lsbinfo[n],hd.faninfo[n]);
       }
*/

       /* 1st headder block write */

             // note that   VLBA mode assume NRZM with parity
             //             MarkIV mode assume NRZL without parity

       if(hd.vlba==1){     // VLBA mode
	      /* write nrzm initial data */
          if(hd.nbits==32)
	          iwrite=fwrite (&hd.io, sizeof(int), 1, fo); // 32bit(track) mode only
          if(hd.nbits==64)
	          iwrite=fwrite (&hd.lo, sizeof(int64_t), 1, fo); // 64bit(track) mode only
          /* VLBA format header starts  with sync block not AUX block */ 
          if(hd.nbits==32){
             m5_hd_synctime_make(&hd);  //header sync and time block only make
             iwrite=fwrite(hd.ihead,sizeof(int),108,fo); // with parity only
          }
          if(hd.nbits==64){
             m5_hd_synctime_make64(&hd);  //header sync and time block only make
             iwrite=fwrite(hd.lhead,sizeof(int64_t),108,fo); // with parity only
          }
       } else {            // MarkIV mode  
          /* Mark-IV format header */
          m5_hd_make_IV(&hd);
          if(hd.nbits==32)
            iwrite=fwrite(hd.ihead,sizeof(int),160,fo); // without parity mode only
          if(hd.nbits==64){
            iwrite=fwrite(hd.lhead,sizeof(int64_t),160,fo); // without parity mode only
          }
       }

       //below is for check
       
       //head_check(&hd);

       //printf("int64_t size  %d\n",sizeof(int64_t));
       //exit(0);

       //monitor debug
       //monit_head(&hd);



       //printf("ifsecdrame %d\n",hd.ifsecframe);

       /* conversion start message */

       printf("\n");
       printf("K5 file(s)\n");
       k=1;  // K5 ch# start
       for(n=0; n<runp.k5num; n++){
           if(k5[n].numch==1){ // 1ch mode data
               printf(" FILE%d CH# %2d    : %s\n",n+1,k, k5[n].fname);
               k++;
           } else {  //4ch mode
               printf(" FILE%d CH# %2d-%2d : %s\n",n+1,k,k+3, k5[n].fname);
               k+=4;
           }
           printf("   (A/D(bits) %d  CHs %d  SFreq(kHz) %d",
                     k5[n].adbit,k5[n].numch,k5[n].sfreq);
           printf(" Time %02d:%02d:%02d  sec %d)\n",
                      k5[n].hh,k5[n].mm,k5[n].ss,k5[n].seconds);
       }

       printf("\n");
       printf("Mark5 file (created) : %s\n",runp.mk5name);
       printf("    Observation Date : 3digit MJD = %4d\n",runp.mjd3);
       if(runp.yyyy != 0){
       printf("                     :      Year  = %4d\n",runp.yyyy);
       printf("                     :      Month = %4d\n",runp.mm);
       printf("                     :      Day   = %4d\n",runp.dd);
       printf("                     :  Total Day = %4d\n",runp.ddd);
       }
       printf("\n");
       printf("                   TIME LABEL MONITOR\n");
       if(hd.vlba==1){
       printf(" ==========================================================\n");   
       printf("  K5-FILE1  K5-FILE2  K5-FILE3  K5-FILE4    VLBA-TIME\n");   
       printf("  HH:MM:SS  HH:MM:SS  HH:MM:SS  HH:MM:SS  JJJSSSSSssss\n");   
       printf(" ----------------------------------------------------------\n");   
       } else {
       printf(" ==========================================================\n");   
       printf("  K5-FILE1  K5-FILE2  K5-FILE3  K5-FILE4   MarkIV-TIME\n");   
       printf("  HH:MM:SS  HH:MM:SS  HH:MM:SS  HH:MM:SS  YDDDHHMMSSsss\n");   
       printf(" ----------------------------------------------------------\n");   
       }

       tmax=k5[0].seconds;
       tmin=tmax;
       for(n=0; n<runp.k5num; n++){
           printf("  %02d:%02d:%02d",k5[n].hh,k5[n].mm,k5[n].ss);
           if(k5[n].seconds > tmax) tmax=k5[n].seconds;
           if(k5[n].seconds < tmin) tmin=k5[n].seconds;
       }
       if(runp.k5num != 4){
           for(n=runp.k5num; n<4; n++) printf("          ");
       }
       if(hd.vlba==1){
           printf("  %03d%05d%04d",hd.mjd,hd.seconds,hd.ifsec);
       } else {
           printf("  %1d%03d%02d%02d%02d%03d",
               (hd.yyyy % 10),hd.ddd,(int)(hd.seconds/3600),
               (int)((hd.seconds%3600)/60),(hd.seconds%60),
               hd.ifsec/10);
       }
       if(tmax==tmin){
           printf("  OK\n");
       } else {
           printf("  ERROR (TIME DIFFER!!)\n");
           exit(1);
       }

       period=0;

       pk=0;
       //for(n=0; n<32; n++) pdat[n]=0;
       for(n=0; n<hd.nbits; n++) pdat[n]=0;
       ow=0;


       koffset=0;
       if(hd.vlba==0){
           koffset=160;
       }

       /*work table set up */
       setup_work_table(fanout, adbit, bittable, &mk5info, &hd);

       for(m=0; m<hd.numch; m++){
           filen_tabl[m]=0;
           k5chn_tabl[m]=0;
           i=0;
           k=0;
           for(n=0; n< runp.k5num; n++){
              i+=k5[n].numch;
              if((m - i) < 0 ){
                  filen_tabl[m]=n;      // this K5 file#-1
                  k5chn_tabl[m]=m-k;    // this is K5 Ch# in a file -1
                  break;
              }
              k=i;
           }
       }

       //for(m=0; m<hd.numch; m++){ // debug
       //  printf("ch# %d filen = %d  k5ch# = %d\n",m,filen_tabl[m],k5chn_tabl[m]);
       //}


       /*
       i=0;
       for(fo_pos=1; fo_pos<=fanout; fo_pos++){
           //for(n=0; n<runp.k5num; n++){
           //for(n=0; n<16; n++){
           for(n=0; n<mk5info.nch; n++){
               m5ch=n+1;  // Mark5 CH#  (1,2,....)
               //printf("m5ch= %d\n",m5ch);
               for(a=0; a<adbit; a++){   // adbit=1   a=0: sign bit
                                         // adbit=2   a=0: mag bit
                                         //           a=1: sign bit
                   adinfo=adbit-a-1; // adinfo  =0: sign bit  =1: mag bit
                   //bitpos=get_bitpos(m5ch,adinfo,fo_pos,&track,&hstack);
                   bitpos=get_bitpos2(m5ch,adinfo,fo_pos,&mk5info,&hd);
                   bittable[(fo_pos-1)+(m5ch-1)*fanout+adinfo*fanout*16]=bitpos;
                   i++;
               }
           }
       }
       */

       //printf("numb = %d\n",numb);


       while(1){



           for(i=0; i<imax; i++){  // 1 sec loop is imax'th Mark5 frame
               for(n=0; n<runp.k5num; n++){
                  iread = fread( k5[n].ibuf, sizeof( char ), numb, k5[n].stream);
                  parseK5(k5[n].ibuf,numb,adbit,numch,k5[n].idata, &kpos);
               }
               owp=koffset;
               // convert to Mark5 
               //for(k=0; k<kpos; k++){
               //printf("Debug %10d %10d %10d\n",i,hd.seconds,hd.ifsec);
               for(kk=koffset; kk<kpos/fanout; kk++){
                   //bitpos=0;
                   iw=0; lw=0;
                   for(fo_pos=1; fo_pos<=fanout; fo_pos++){
                       k=fanout*kk+fo_pos-1;

                       /* following 't' is jusr for generating dummy sine data */ 
                       //t=(double)hd.seconds+(double)hd.ifsec/10000.0
                       //    +(double)k/fsampl;   // for seconds in a day
                       t=(double)hd.ifsec/10000.0
                           +(double)k/fsampl;   // for fractional second only
                       //printf("  Debug %5d %5d %25.15f\n",kk,fo_pos,t);

                       for(ch=0; ch < hd.numch; ch++){ 
                           // ch should be channel order in K5 data
                           m5ch=mk5info.ch_ch[ch];   // Mark5 channel
                           // get k5 file number (starting from 0) including m5ch data
                           // ch#(starting from 0) in a K5 file

                           /* get K5 file number and K5 ch# */
                           n=filen_tabl[ch];
                           m=k5chn_tabl[ch];
//                           iok=m5ch_ok_tabl[m5ch-1]; // now all ok
//                           if(iok==1){
                              /*
                              put_k5_data(adbit,m5ch,fanout,fo_pos,k,m,
                                            pdat,bittable,&iw,set_table,
                                            &lw,set_ltable,&k5[n],&hd);
                              */
                              // following code is faster than call above function
                              for(a=0; a<adbit; a++){   // adbit=1   a=0: sign bit
                                                     // adbit=2   a=0: mag bit
                                                     //           a=1: sign bit
                                adinfo=adbit-a-1; // adinfo  =0: sign bit  =1: mag bit
                                sdat=k5[n].idata[m+a*numch+k*numch*adbit];
                                if(sdat==1){ //set data at bitpos
                                   //bitpos get
                                   //bitpos=get_bitpos(m5ch,adinfo,fo_pos,&track,&hstack);
                                   bitpos=bittable[(fo_pos-1)+(m5ch-1)*fanout+adinfo*fanout*16];
                                   //if(bitpos >=64) printf(" bitpos %d",bitpos);//debug
                                   if(hd.nbits==64){
                                       lw |= set_ltable[bitpos];
                                   } else {
                                       iw |= set_table[bitpos];
                                   }
                                   pdat[bitpos]++;  // count # of 1 for parity
                                } 
                             }
                             
//                           } else {  //put 1MHz dummy sine signal  (usuallu not used)
//                              put_sine_wave(t,adbit,m5ch,fanout,fo_pos,
//                                            pdat,bittable,&iw,set_table,
//                                            &lw,set_ltable,&hd);
//                           }
                       }
                   }

                   if(hd.vlba==1){
                       if(hd.nbits==64){
                           hd.lo^= lw;
                           lbuf[ow]=hd.lo;
                       } else {
                           hd.io ^= iw;
                           obuf[ow]=hd.io;
                       }
                   } else {
                       if(hd.nbits==64){
                           lbuf[ow]=lw;
                       } else {
                           obuf[ow]=iw;
                       }
                   }
                   //iwrite=fwrite(&hd.io,sizeof(int),1,fo); 

                   pk++;
                   ow++;
                   owp++;  // for check
                   if(hd.vlba==1 && pk==8) { //parity set calc
                       iw=0; lw=0;
                       //for(n=0; n<32; n++){
                       if(hd.nbits==64){
                           for(n=0; n<64; n++){
                               if((pdat[n] % 2) == 0){  // odd parity
                                   lw |= set_ltable[n];
                               }
                           }
                           hd.lo ^= lw;
                           //obuf[pk]=hd.io;
                           lbuf[ow]=hd.lo; 
                       } else {
                           for(n=0; n<32; n++){
                               if((pdat[n] % 2) == 0){  // odd parity
                                   iw |= set_table[n];
                               }
                           }
                           hd.io ^= iw;
                           //obuf[pk]=hd.io;
                           obuf[ow]=hd.io; 
                       }
                       ow++;

                       for(n=0; n<hd.nbits; n++) pdat[n]=0;

                       pk=0;

                   }

                   if(ow==obufmax){
                       if(hd.nbits==64){
                           iwrite=fwrite(lbuf,sizeof(int64_t),obufmax,fo);
                       } else {
                           iwrite=fwrite(obuf,sizeof(int),obufmax,fo);
                       }
                       ow=0;
                   }
               }
               if(ow!=0){
                   if(hd.nbits==64){
                        iwrite=fwrite(lbuf,sizeof(int64_t),ow,fo);
                   } else {
                        iwrite=fwrite(obuf,sizeof(int),ow,fo);
                   }
                   ow=0;
               }

               //printf("owp= %d\n",owp);

//debug
/*
               for(n=0; n<runp.k5num; n++){
                   printf("FILE # %d\n",n+1);
                   mon_parseK5(k5[n].idata, adbit, numch);
               }
               exit(0);
*/
               if(iread < numb) goto exit_loop;  // exit loop

               /* insert full header */
               hd.ifsec+=hd.ifsecframe;
               if(hd.ifsec >= 10000){
                  hd.ifsec-=10000;
                  hd.seconds++;
                  if(hd.seconds >= 86400){
                      hd.seconds-=86400;
                      hd.mjd++;
                  }
               } 
               if(hd.vlba==1){
                   if(hd.nbits==64){
                      m5_hd_make64(&hd);  // Mark5 header generate
                      iwrite=fwrite(hd.lhead,sizeof(int64_t),180,fo); // with parity only
                   } else {
                      m5_hd_make(&hd);  // Mark5 header generate
                      iwrite=fwrite(hd.ihead,sizeof(int),180,fo); // with parity only
                   }
               } else {
                   m5_hd_make_IV(&hd);  // Mark4 header generate
                   if(hd.nbits==32)
                      iwrite=fwrite(hd.ihead,sizeof(int),160,fo); // without parity mode only
                   if(hd.nbits==64)
                      iwrite=fwrite(hd.lhead,sizeof(int64_t),160,fo); // without parity mode only
               }
           }

           // second data read end

           period++;
           if(runp.period > 0 && (runp.period == period)) goto exit_loop;
           // K5 header read
           tmax=k5[0].seconds;
           tmin=tmax;
           for(n=0; n<runp.k5num; n++){
              if(chk_K5_header(&k5[n]) < 0){
                 goto exit_loop;
              }
              if(n==0){
                 tmax=k5[0].seconds;
                 tmin=tmax;
              }
              printf("  %02d:%02d:%02d",k5[n].hh,k5[n].mm,k5[n].ss);
              if(k5[n].seconds > tmax) tmax=k5[n].seconds;
              if(k5[n].seconds < tmin) tmin=k5[n].seconds;
           } 
           if(runp.k5num != 4){
              for(n=runp.k5num; n<4; n++) printf("          ");
           }
           if(hd.vlba==1){
               printf("  %03d%05d%04d",hd.mjd,hd.seconds,hd.ifsec);
           } else {
               printf("  %1d%03d%02d%02d%02d%03d",
                   (hd.yyyy % 10),hd.ddd,(int)(hd.seconds/3600),
                   (int)((hd.seconds%3600)/60),(hd.seconds%60),
                   hd.ifsec/10);
           }
           if(hd.ifsec != 0){
              printf("  ERROR (TIME DIFFER!!)\n");
           } else {
              if(tmax==tmin){
                  if(hd.seconds==k5[0].seconds){
                     printf("  OK\n");
                  } else {
                      printf("  ERROR (TIME DIFFER!!)\n");
                  }
              } else {
                 printf("  ERROR (TIME DIFFER!!)\n");
              }
           }
       }

exit_loop:

       printf(" ==========================================================\n");   

       for(n=0; n<runp.k5num; n++) i=fclose(k5[n].stream);

       if(ow!=0){
           if(hd.nbits==64){
               iwrite=fwrite(lbuf,sizeof(int64_t),ow,fo);
           } else {
               iwrite=fwrite(obuf,sizeof(int),ow,fo);
           }
       }

       i=fclose(fo);


       printf("\n");
       printf("  Mark5 file (%s) has been created.\n\n",runp.mk5name);




       tfinish= time(NULL);
       telapsed = difftime(tfinish,tstart);
       printf("Time elapsed (sec) for One obs process is %f\n",telapsed);

exitend: 
       return 0;
}
