/*************************************************************************/
/*                                                                       */
/*   Program Name :  m5time                                              */
/*                                                                       */
/*     M5time monitors header of Mark-5 file                             */
/*                                                                       */
/*     Enviroment variables                                              */
/*           M5DIR      ---   default directory for Mark5 data           */
/*                                                                       */
/*      Version:  1.10  2005-06-17                                       */
/*                                                                       */
/*  copyright (c) 2003-2005 T.Kondo/NICT(CRL) All Right Reserved         */
/*                                                                       */
/*************************************************************************/
/*

  Mark-5 data sync block monitor

          T.Kondo 2004.11.13
                  2005.06-17 update due to the library restructuring
                 

    m5time mk5_filename [options]
    
     where   mk5filename -- Mark 5 data file name
              options (any order)
               -b bitpos  -- pickup bitposition (0- (ntrack-1))
                            (default 0)
               -t[rack] ntrack -- #of bits/word
                            (default 32)
               -n nkai -- # of frames to display (read)
                            (default 20)
               -vlba      -- file is VLBA Mark5 data

G:\mark5\m5test.dat -b 0 -vlba -t 64
G:\IPVLBI\data\11-K178-1321 -b 0 -vlba -t 32
G:\GSICHECK\K04242\242-0734 -b 2
*/
#define PROGRAM_VERSION "m5time (Ver 1.10  2005-06-17)"


#include "libk5.h"      // common header for IP-VLBI utilities 

#include "libmk5tok5.h"  // library for Mark5 to K5 conversion

/* environment variable for Mark5 data default directory set */
#define M5DIR "M5DIR" // the environment variable for Mark5 data default directory


void param_get_m5time(int argc, char *argv[],
               char *mk5name, 
               int *bitpos, int *ntrack, int *nkai, int *vlba)
{
    int n;
    char cbuf[80];
    //1st param should be mark5 name
    if (argc>=2){
       strcpy(mk5name,*(argv+1));
    } else {
       mk5name[0]='\0';
    }

    /* check Mark5 file name is acompanied by directory or not */
    get_true_path(mk5name, M5DIR, "./",cbuf);
    strcpy(mk5name,cbuf);

    // initialize parameters
    *bitpos=0;
    *ntrack=32;
    *nkai=20;
    *vlba=0;

    n=3;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));

        if(strncmp(cbuf,"-b",2)==0){  // bitpos
            n++;
            if(n > argc) break;
            *bitpos = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-t",2)==0){  // ntrack
            n++;
            if(n > argc) break;
            *ntrack = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-n",2)==0){  // nkai
            n++;
            if(n > argc) break;
            *nkai = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-v",2)==0){  // VLBA mode
            *vlba=1;
        }
        n++;
    }
}

void print_usage(void)
{
    //printf("m5tok5Rnp  %s\n",M5TOK5RNP_VERSION);
    printf("%s\n",PROGRAM_VERSION);
    printf("Usage: m5time mk5_filename [options]\n");
    printf("\n");
    printf("where   mk5filename -- Mark5 data file name\n");
    printf("       options (any order)\n");
    printf("        -b bitpos  -- pickup bitposition (0- (ntrack-1))\n");
    printf("                      (default 0)\n");
    printf("        -t[rack] ntrack -- #of bits/word\n");
    printf("                          (default 32)\n");
    printf("        -n nkai -- # of frames to display (read)\n");
    printf("                            (default 20)\n");
    printf("        -v[lba]      -- file is VLBA Mark5 data\n");
    printf("   Enviroment variables\n");
    printf("        %s    ---   default directory for Mark5 data\n",M5DIR);
    printf("\n");
}


int main(int argc, char *argv[])
{
       int i,n;
       int bitpos,ntrack,vlba,nkai;
       char mk5name[80];
       char program_version[60];
       char* dflt_fname="G:/IPVLBI/data/jiveexp/Wb_test";
      // IPBOARD_T ipb;
       mk5_data_t mk5;    // to suport 2 bit sampling 4 ch and fanout 4 conversion

       if(argc == 1){
          print_usage();
          exit(0);
       }
       if (argc>=2){
          if(strncmp(*(argv+1),"env",3)==0){
          printf("   Enviroment variables\n");
          printf("        %s ---   default directory for Mark5 data\n",M5DIR);
          printf("                      ( %s )\n",getenv(M5DIR));
          printf("\n");
          exit(0);
          }
       }

       strcpy(program_version,PROGRAM_VERSION);

       /* Program Name Display */
       printf("****************************************************\n");
       printf("*    Mark-5 data time monitor                      *\n");
       printf("*      %s               *\n",PROGRAM_VERSION);
       printf("*          Kashima Space Research Center/NICT      *\n");
       printf("****************************************************\n");
       printf("\n");

       /* run parameter get */
       param_get_m5time(argc, argv, mk5name, &bitpos, &ntrack, &nkai, &vlba);

       if (mk5name[0]=='\0'){
          strcpy(mk5name,dflt_fname);
       }

       /* Mark5 file existence check */
       if( (mk5.fin  = fopen( mk5name, "rb" )) == NULL ){
           printf("Mark5 File %s   Open Error! \n",mk5name);
		   goto exitend;
       }
       i=fclose(mk5.fin);



       printf("Mark 5 Data File : %s\n",mk5name);

       printf("ntrack= %d  bitpos = %d\n",ntrack,bitpos);
       if(vlba==1){
           mk5.vlba=1;
           mk5.nrzm=1;
           mk5.parity=1;
           printf("VLBA mode\n");
       } else {
           mk5.vlba=0;
           mk5.nrzm=0;
           mk5.parity=0;
           printf("Mark4 mode\n");
       }
       if(vlba==0){ // Mark-IV format
 	       printf("       HEADER     AUX     SYNC    TIME FIELD  CRC  \n");
 	       printf("    # aaaabbbb rrssttuu          YDDDHHMM SSsss 12\n");
       } else {  // VLBA format
           printf("       HEADER     AUX     SYNC    TIME CODE   CRC \n");
 	       printf("    # SSRHPPPP TGBSCD00          MJDsssss.ssss  16\n");
       } 
       mk5.bitpos=bitpos;
       mk5.ko=0;
       mk5.ntrack=ntrack;

           /* Data file open */
       if( (mk5.fin  = fopen( mk5name, "rb" )) == NULL ){
              printf("Data File %s   Open Error! \n",mk5name);
		      goto exitend;
       }


       for(n=0; n<nkai; n++){
          //if(vlba==1){
          //    i=read_1frame(&mk5);
          //} else {
          //    i=read_1frame_np(&mk5); 
          //}
          i=read_1frame_r(&mk5);

	      if(i<=0) {
		     i=fclose(mk5.fin);
		     goto exitend;
		  }

          printf("%5d",n);
          for (i=0; i<5; i++){   //check           
            printf(" %08X",rvsint(mk5.header[i]));
          } 
          printf("\n");
          //printf("%d %d %d\n",n,mk5.aux.trk,mk5.aux.adid); //debug
       
       }
	   
	   i=fclose(mk5.fin);

exitend: 
       return 0;
	   
}
